//! Common dnssec utils

use ::std::{net::SocketAddr, vec::Vec};
use ::tracing::error;
use ::trust_dns_resolver::TokioAsyncResolver;

pub mod record;
pub use record::Record;

/// Common errors for Dnssec setup and usage
#[derive(::thiserror::Error, Debug)]
pub enum Error {
    /// Error reading or parsing /etc/resolv.conf
    #[error("resolv.conf: {0:?}")]
    ResolvConf(String),
    /// Error in the resolver setup
    #[error("resolver setup: {0:?}")]
    Setup(String),
    /// Errors in establishing connection or  connection drops
    #[error("nameserver connection: {0:?}")]
    NameserverConnection(String),
}

#[cfg(any(
    target_os = "linux",
    target_os = "macos",
    target_os = "freebsd",
    target_os = "dragonfly",
    target_os = "openbsd",
    target_os = "netbsd",
))]

/// Easy Dnssec handling for Fenrir
#[allow(missing_debug_implementations)]
pub struct Dnssec {
    /// real resolver
    resolver: TokioAsyncResolver,
}

impl Dnssec {
    /// Spawn connections to DNS via TCP
    pub async fn new(resolvers: &Vec<SocketAddr>) -> Result<Self, Error> {
        // use a TCP connection to the DNS.
        // the records we need are big, will not fit in a UDP packet
        let resolv_conf_resolvers: Vec<SocketAddr>;
        let resolvers = if resolvers.len() > 0 {
            resolvers
        } else {
            // load from system's /etc/resolv.conf
            let resolv_conf =
                match ::trust_dns_resolver::system_conf::read_system_conf() {
                    Ok((resolv_conf, _)) => resolv_conf,
                    Err(e) => return Err(Error::ResolvConf(e.to_string())),
                };
            resolv_conf_resolvers = resolv_conf
                .name_servers()
                .iter()
                .map(|r| r.socket_addr)
                .collect();

            &resolv_conf_resolvers
        };
        use ::trust_dns_resolver::config::{ResolverConfig, ResolverOpts};
        let mut opts = ResolverOpts::default();
        opts.edns0 = true;
        opts.validate = true;
        opts.use_hosts_file = false;
        opts.try_tcp_on_error = true;
        let opts = opts; // no more mut

        // default() uses google's dns: don't use
        let mut config = ResolverConfig::new();

        for resolver in resolvers.iter() {
            use ::trust_dns_resolver::config::{NameServerConfig, Protocol};
            config.add_name_server(NameServerConfig::new(
                resolver.clone(),
                Protocol::Tcp,
            ));
        }

        let resolver = match TokioAsyncResolver::tokio(config, opts) {
            Ok(resolver) => resolver,
            Err(e) => return Err(Error::Setup(e.to_string())),
        };

        Ok(Self { resolver })
    }
    /// Get the fenrir data for a domain
    pub async fn resolv(&self, domain: &str) -> ::std::io::Result<Record> {
        use ::trust_dns_client::rr::Name;

        let fqdn_str = "_fenrir.".to_owned() + domain;
        ::tracing::debug!("Resolving: {}", fqdn_str);
        let fqdn = Name::from_utf8(&fqdn_str)?;
        let answers = self.resolver.txt_lookup(fqdn).await?;
        const TXT_RECORD_START: &str = "v=Fenrir1 ";
        let mut found_but_wrong: bool = false;
        for txt_raw in answers.into_iter() {
            let txt = ::std::format!("{}", txt_raw);
            if !txt.starts_with(TXT_RECORD_START) {
                ::tracing::trace!("Found NON-fenrir record: {}", txt);
                continue;
            }
            ::tracing::trace!("Found fenrir record: {}", txt);
            let base85 = txt[TXT_RECORD_START.len()..].trim();
            let bytes = match ::base85::decode(base85) {
                Ok(bytes) => bytes,
                Err(e) => {
                    ::tracing::error!("Invalid DNSSEC base85: {}", e);
                    continue;
                }
            };
            let record = match Record::decode(&bytes) {
                Ok(record) => record,
                Err(e) => {
                    found_but_wrong = true;
                    ::tracing::error!("{}: {}", fqdn_str, e);
                    continue;
                }
            };
            return Ok(record);
        }
        return Err(if found_but_wrong {
            ::std::io::Error::new(
                ::std::io::ErrorKind::InvalidData,
                "record found but not parsable",
            )
        } else {
            ::std::io::Error::new(
                ::std::io::ErrorKind::NotFound,
                "record not found",
            )
        });
    }
}
